#!/usr/bin/env python3
"""Trufflehog3 CLI."""

import argparse
import git
import logging
import multiprocessing
import os
import sys

from pathlib import Path
from signal import signal, SIGINT
from tempfile import TemporaryDirectory
from typing import Callable
from urllib.parse import urlparse

from trufflehog3 import __NAME__, __VERSION__
from trufflehog3 import DEFAULT_RULES_FILE
from trufflehog3 import log

from trufflehog3.core import diff, load, load_config, load_rules, render, scan
from trufflehog3.models import (
    Config,
    Exclude,
    Format,
    Issue,
    Severity,
)

MORE = f"""
learn more:
  https://github.com/feeltheajf/{__NAME__}

version:
  {__VERSION__}
"""

try:
    CPU_COUNT = multiprocessing.cpu_count()
except NotImplementedError:  # pragma: no cover
    CPU_COUNT = 1


def run(**kwargs):
    """Run CLI."""
    args = _get_cmdline_args(**kwargs)
    log.setLevel(logging.ERROR - args.verbose * 10)

    if args.version:  # pragma: no cover
        print(__VERSION__)
        return 0

    if args.render_html:  # pragma: no cover
        issues = []
        for f in args.targets:
            issues.extend(load(Issue, f))
        render(issues, format=Format.HTML, file=args.output)
        return 0

    kw = {k: v for k, v in args.__dict__.items() if hasattr(Config(), k) and v}
    if args.config:  # pragma: no cover
        config = load_config(args.config, **kw)

    rules = load_rules(args.rules, args.severity)
    issues = []

    for target in args.targets:
        remote = urlparse(target).scheme in ("http", "https")
        if remote:  # pragma: no cover
            tmp = TemporaryDirectory(prefix=f"{__NAME__}-")
            git.Repo.clone_from(target, tmp.name)
            target = tmp.name

        if not args.config:
            config = load_config(target, **kw)

        issues.extend(scan(target, config, rules, args.processes))

        if remote:  # pragma: no cover
            tmp.cleanup()

    if args.incremental:  # pragma: no cover
        issues = diff(load(Issue, args.incremental), issues, only_new=True)

    render(issues, format=args.format, file=args.output)
    return 0 if args.zero else 2 if issues else 0


class _HelpFormatter(argparse.RawTextHelpFormatter):  # pragma: no cover
    def __init__(self, prog):
        super().__init__(prog, max_help_position=26)

    def _format_action_invocation(self, action):
        if not action.option_strings or action.nargs == 0:
            return super()._format_action_invocation(action)

        options = ", ".join(action.option_strings)
        if action.choices:
            options += " {%s}" % ", ".join(str(c) for c in action.choices)
        elif action.metavar:
            options += " " + action.metavar

        return options


def _exclude(s: str) -> Exclude:
    """Convert string to exclude rule.

    Examples
    --------
    Basic usage examples

    >>> _exclude("p1,p2")
    Exclude(message='p1,p2', id=None, pattern=None, paths=['p1', 'p2'])

    >>> _exclude("re:p")
    Exclude(message='re:p', id=None, pattern=re.compile('re'), paths=['p'])
    """
    if ":" in s:
        pattern, paths = s.split(":", 1)
    else:
        pattern, paths = None, s
    return Exclude(message=s, pattern=pattern, paths=paths.split(","))


def _file(mode: str = "r") -> Callable[[str], Path]:  # pragma: no cover
    def validate(filepath: str) -> Path:
        path = Path(filepath)
        file = path.open(mode)
        file.close()
        return path

    return validate


def _get_cmdline_args(**defaults) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Find secrets in your codebase",
        epilog=MORE,
        usage=f"{__NAME__} [arguments] targets",
        formatter_class=_HelpFormatter,
    )
    parser.add_argument(
        "targets",
        help="Search targets, defaults to current directory",
        nargs=argparse.ZERO_OR_MORE,
        default=[os.curdir],
    )
    parser.add_argument(
        "-z",
        "--zero",
        help="always exit with zero status code",
        dest="zero",
        action="store_true",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        help="enable verbose logging {-v, -vv, -vvv}",
        dest="verbose",
        action="count",
        default=0,
    )
    parser.add_argument(
        "-o",
        "--output",
        help="path to output file",
        dest="output",
        metavar="file",
        type=_file("w"),
    )
    parser.add_argument(
        "-c",
        "--config",
        help="path to config file",
        dest="config",
        metavar="file",
        type=_file("r"),
    )
    parser.add_argument(
        "-r",
        "--rules",
        help="path to rules file",
        dest="rules",
        metavar="file",
        type=_file("r"),
        default=DEFAULT_RULES_FILE,
    )
    parser.add_argument(
        "-i",
        "--incremental",
        help="path to previous scan",
        dest="incremental",
        metavar="file",
        type=_file("r"),
    )
    parser.add_argument(
        "-p",
        "--processes",
        help="number of subprocesses to run (%(default)s)",
        dest="processes",
        metavar="int",
        type=int,
        default=CPU_COUNT,
    )
    search = parser.add_argument_group("search arguments")
    search.add_argument(
        "-e",
        "--exclude",
        help="exclude matching issues",
        dest="exclude",
        metavar="str",
        type=_exclude,
        nargs=argparse.ZERO_OR_MORE,
    )
    search.add_argument(
        "-s",
        "--severity",
        help="minimum severity filter (%(default)s)",
        dest="severity",
        metavar="str",
        type=Severity,
        choices=[Severity.LOW, Severity.MEDIUM, Severity.HIGH],
        default=Severity.LOW,
    )
    search.add_argument(
        "--ignore-nosecret",
        help="ignore inline 'nosecret' annotations",
        dest="ignore_nosecret",
        action="store_true",
    )
    mgroup = search.add_mutually_exclusive_group()
    mgroup.add_argument(
        "--no-entropy",
        help="disable entropy checks",
        dest="no_entropy",
        action="store_true",
    )
    mgroup.add_argument(
        "--no-pattern",
        help="disable pattern checks",
        dest="no_pattern",
        action="store_true",
    )
    source = parser.add_argument_group("source arguments")
    source.add_argument(
        "--branch",
        help="name of the repo branch to scan",
        dest="branch",
        metavar="str",
    )
    commit = source.add_mutually_exclusive_group()
    commit.add_argument(
        "--depth",
        help="max commits depth for searching (%(default)s)",
        dest="depth",
        type=int,
        metavar="int",
        default=10000,
    )
    commit.add_argument(
        "--since",
        help="scan from the given commit hash",
        dest="since",
    )
    mgroup = source.add_mutually_exclusive_group()
    mgroup.add_argument(
        "--no-current",
        help="disable current status check",
        dest="no_current",
        action="store_true",
    )
    mgroup.add_argument(
        "--no-history",
        help="disable commit history check",
        dest="no_history",
        action="store_true",
    )
    render = parser.add_argument_group("render arguments")
    render.add_argument(
        "-f",
        "--format",
        help="output format (%(default)s)",
        dest="format",
        metavar="str",
        type=Format,
        choices=[Format.TEXT, Format.JSON, Format.HTML],
        default=Format.TEXT,
    )
    render.add_argument(
        "--context",
        help="number of context lines to include",
        dest="context",
        metavar="int",
        type=int,
        default=0,
    )
    helper = parser.add_argument_group("other commands")
    others = helper.add_mutually_exclusive_group()
    others.add_argument(
        "-R",
        "--render-html",
        help="render HTML report from JSON",
        dest="render_html",
        action="store_true",
    )
    others.add_argument(
        "-V",
        "--version",
        help="show version message and exit",
        action="store_true",
    )
    parser.set_defaults(**defaults)
    return parser.parse_args()


def _exit_on_keyboard_interrupt(*args, **kwargs):  # pragma: no cover
    sys.stdout.write("\r")
    log.fatal("Keyboard interrupt. Exiting")
    sys.exit(0)


if __name__ == "__main__":  # pragma: no cover
    signal(SIGINT, _exit_on_keyboard_interrupt)
    sys.exit(run())



"""Helper classes for passing data around."""

import attr
import re
import string
import uuid

from abc import ABC, abstractmethod
from datetime import datetime
from enum import auto, Enum, EnumMeta
from pathlib import Path
from typing import Any, Dict, List, Optional

from trufflehog3 import log, helper, IGNORE_NOSECRET

_NAMESPACE = uuid.UUID("00000000-0000-0000-0000-000000000000")

BASE64_CHARS = string.ascii_letters + string.digits + "+/="
BASE64_LIMIT = 4.5
HEX_CHARS = string.hexdigits
HEX_LIMIT = 3.0


class CaseInsensitiveEnumMeta(EnumMeta):
    """Meta class for case-insensitive enum."""

    def __call__(self, value, *args, **kwargs):
        """Search enum names instead of values when value is string."""
        if isinstance(value, str):
            return self.__getitem__(value)

        return super().__call__(value, *args, **kwargs)

    def __getitem__(self, name):
        """Make get item operation case-insensitive."""
        return super().__getitem__(name.upper())


class Severity(Enum, metaclass=CaseInsensitiveEnumMeta):
    """Issue severity based on match confidence and other factors."""

    LOW = auto()
    MEDIUM = auto()
    HIGH = auto()

    def __lt__(self, other):  # pragma: no cover
        """Severity comparison is used for sorting issues."""
        return self.value < other.value

    def __le__(self, other):
        """Severity equality is used for sorting issues."""
        return self.value <= other.value

    def __str__(self):
        """Override string method to return enum name."""
        return self.name

    def __neg__(self):
        """Override negation method to use enum value."""
        return -self.value


class Format(Enum, metaclass=CaseInsensitiveEnumMeta):
    """Supported output formats."""

    TEXT = auto()
    JSON = auto()
    HTML = auto()

    def __str__(self):  # pragma: no cover
        """Override string method to return enum name."""
        return self.name


@attr.s
class Model:
    """Model is a base class for all models definitions."""

    def asdict(self):
        """Convert model to dictionary."""
        return attr.asdict(
            self, filter=lambda attr, _: not attr.name.startswith("_")
        )


@attr.s(frozen=True)
class File(Model):
    """File is a basic wrapper with Git metadata support.

    Attributes
    ----------
    path (str)
    : File path.

    branch (str, optional)
    : Git commit branch.

    message (str, optional)
    : Git commit message.

    author (str, optional)
    : Git commit author as `name <email>`.

    commit (str, optional)
    : Git commit hash.

    date (datetime.datetime, optional)
    : Git commit timestamp.

    Args
    ----
    content (str, optional)
    : File content.

    Examples
    --------
    Basic usage examples

    >>> f = File("tests/data/test_file.txt")
    >>> f.read()
    'Test'
    >>> f = File("nosuchpath/test_file.txt", content="Test")
    >>> f.read()
    'Test'

    """

    path: str = attr.ib()
    branch: Optional[str] = attr.ib(None)
    message: Optional[str] = attr.ib(None)
    commit: Optional[str] = attr.ib(None)
    author: Optional[str] = attr.ib(None)
    date: Optional[datetime] = attr.ib(None)
    _content: Optional[str] = attr.ib(None)
    _real: Optional[str] = attr.ib(None)

    def read(self) -> str:
        """Return the given content or read file from path."""
        if self._content is not None:
            return self._content

        try:
            return Path(self._real or self.path).read_text()
        except Exception as e:  # pragma: no cover
            log.warning(f"skipping file '{self.path}': {e}")
            return ""


@attr.s
class Rule(Model, ABC):
    """Rule is a base class for rules definitions."""

    def __eq__(self, other):  # pragma: no cover
        """Override equality check to use rule IDs."""
        return (type(self) == type(other)) and (self.id == other.id)

    def __hash__(self):  # pragma: no cover
        """Override hash value to use rule ID."""
        return hash(self.id)

    @abstractmethod
    def findall(self, s: str) -> List[str]:  # pragma: no cover
        """Find all substrings matching rule."""
        ...

    @staticmethod
    def fromany(x: Any) -> Any:
        """Convert any object to rule."""
        return x if issubclass(type(x), Rule) else Rule.fromdict(x)

    @staticmethod
    def fromdict(x: Dict[str, Any]) -> Any:
        """Convert dict to rule subclass."""
        try:
            return Pattern(**x)
        except Exception:
            return Entropy(**x)

    @staticmethod
    def fromargs(**x: Any) -> Any:
        """Convert args to rule subclass."""
        return Rule.fromdict(x)


@attr.s(frozen=True)
class Entropy(Rule):
    """Entropy is used for detecting high entropy strings.

    Attributes
    ----------
    id (str)
    : Rule ID, should be unique.

    message (str)
    : Short explanation of what is matched by the rule.

    severity (Severity)
    : Severity of issues detected by the rule.

    Args
    ----
    alphabet (str, optional)
    : Alphabet to search characters from.

    threshold (float, optional)
    : Shannon entropy threshold.

    minlen (int, optional)
    : Minimum match length.

    Examples
    --------
    There are two ways to customize high entropy check.
    The easiest one is to set custom minimum length for matched strings.
    The other way is to set custom alphabets and/or thresholds for them.

    >>> BASE32_CHARS = string.ascii_letters + "234567="
    >>> rule = Entropy(
    ...     alphabet=BASE32_CHARS,
    ...     threshold=3.75,
    ...     minlen=10,
    ... )
    >>> rule.findall("password = 'irtksdajfhaeu356'")
    ['irtksdajfhaeu356']

    """

    id: str = attr.ib("high-entropy")
    message: str = attr.ib("High Entropy")
    severity: Optional[Severity] = attr.ib(Severity.MEDIUM, converter=Severity)
    _alphabet: Optional[str] = attr.ib(BASE64_CHARS)
    _threshold: Optional[float] = attr.ib(BASE64_LIMIT)
    _minlen: Optional[int] = attr.ib(20)
    _uuid: uuid.UUID = attr.ib(init=False)

    @_uuid.default
    def _uuid_default(self):
        return uuid.uuid3(_NAMESPACE, self.id)

    def findall(self, s: str) -> List[str]:
        """Find high entropy substring occurrences in the string.

        Examples
        --------
        Basic usage examples. The first match here is from base64 alphabet
        and the second one exceeded defined hexadecimal entropy threshold.

        >>> rule = Entropy()
        >>> rule.findall("token = 'abcdefghijklmnopqrstuvwxyz'")
        ['abcdefghijklmnopqrstuvwxyz']

        >>> rule = Entropy(
        ...     alphabet=HEX_CHARS,
        ...     threshold=HEX_LIMIT,
        ...     minlen=10,
        ... )
        >>> rule.findall("password = '1234567890'")
        ['1234567890']

        """
        matched = []

        for word in helper.get_strings(s, self._alphabet, self._minlen):
            if helper.shannon_entropy(word, self._alphabet) > self._threshold:
                matched.append(word)

        return matched


@attr.s(frozen=True)
class Pattern(Rule):
    """Pattern holds all neccessary metadata for pattern-based rule definition.

    Attributes
    ----------
    id (str)
    : Rule ID, should be unique.

    message (str)
    : Short explanation of what is matched by the rule.

    pattern (Pattern)
    : Python `re.Pattern` to search for.

    severity (Severity, optional)
    : Severity of issues detected by the rule.

    Examples
    --------
    Match `letmein` string everywhere

    >>> rule = Pattern(
    ...     id="bad-password-letmein",
    ...     message="Bad Password 'letmein'",
    ...     pattern="letmein",
    ...     severity="high",
    ... )

    Match `letmein` Python `re.Pattern`, case-insensitive

    >>> rule = Pattern(
    ...     id="bad-password-letmein",
    ...     message="Bad Password 'letmein'",
    ...     pattern="(?i)letmein",
    ...     severity="high",
    ... )

    """

    id: str = attr.ib()
    message: str = attr.ib()
    pattern: str = attr.ib()
    severity: Optional[Severity] = attr.ib(Severity.MEDIUM, converter=Severity)
    _uuid: uuid.UUID = attr.ib(init=False)
    _pattern: re.Pattern = attr.ib(init=False)

    @_uuid.default
    def _uuid_default(self):
        return uuid.uuid3(_NAMESPACE, self.id)

    @_pattern.default
    def _pattern_default(self):
        return re.compile(self.pattern)

    def findall(self, s: str) -> List[str]:
        """Find pattern occurrences in the string.

        Examples
        --------
        Basic usage examples

        >>> rule = Pattern(
        ...     id="bad-password-letmein",
        ...     message="Bad Password 'letmein'",
        ...     pattern="letmein",
        ...     severity="high",
        ... )
        >>> rule.findall("password = 'letmein'")
        ['letmein']

        """
        return [m.group() for m in self._pattern.finditer(s)]


@attr.s(frozen=True)
class Exclude(Rule):
    """Exclude is used for referencing rules in configuration exclude list.

    Attributes
    ----------
    message (str)
    : Short explanation of what is matched by the rule.

    id (str, optional)
    : Rule ID, should be unique.

    pattern (Pattern, optional)
    : Python `re.Pattern` to search for.

    paths (List[str], optional)
    : File paths for rule to be applied on, defaults to everywhere.
      Each item should be a glob pattern as recognized by `pathlib.Path.match`.

    Note
    ----
    Only one of `id`, `pattern` must be set.

    If `paths` is set, exclude will only be applied on the specified paths.

    Examples
    --------
    Skip rule by its ID, but only in YAML files

    >>> rule = Exclude(
    ...     id="bad-password.letmein",
    ...     message="Not Password 'letmein'",
    ...     paths=["*.yaml", "*.yml"],
    ... )

    Skip lines containing `letmein` string

    >>> rule = Exclude(
    ...     message="Not Password 'letmein'",
    ...     pattern="letmein",
    ... )

    Skip lines containing `letmein` Python `re.Pattern`, case-insensitive

    >>> rule = Exclude(
    ...     message="Not Password 'letmein'",
    ...     pattern="(?i)letmein",
    ... )

    """

    message: str = attr.ib()
    id: Optional[str] = attr.ib(None)
    pattern: Optional[re.Pattern] = attr.ib(
        None, converter=lambda x: re.compile(x) if x else None
    )
    paths: Optional[List[str]] = attr.ib(None)

    def findall(self, s: str) -> List[str]:
        """Find pattern occurrences in the string."""
        return (
            [m.group() for m in self.pattern.finditer(s)]
            if self.pattern
            else []
        )

    @staticmethod
    def fromany(x: Any) -> Any:
        """Convert any object to exclude rule."""
        return x if isinstance(x, Exclude) else Exclude.fromdict(x)

    @staticmethod
    def fromdict(x: Dict[str, Any]) -> Any:
        """Convert dict to exclude rule."""
        return Exclude(**x)


class Context(dict):
    """Dumb workaround for dict being unhashable by default.

    Note
    ----
    It is only intended to be used as `context` property for `models.Issue`,
    which uses its own hashing algorithm.

    """

    __slots__ = ()

    def __eq__(self, other):  # pragma: no cover
        """Return false."""
        return False

    def __hash__(self):  # pragma: no cover
        """Return zero."""
        return 0


@attr.s(frozen=True)
class Issue(Model):
    r"""Issue holds finding metadata.

    Attributes
    ----------
    rule (Rule):
    : Rule the issue was detected by.

    path (str):
    : File path.

    line (int):
    : Line number of the matched line.
      For Git history this is the line number in diff blob, not the real one.

    secret (str):
    : String matched by the rule.

    context (str):
    : Code lines containing secret matched by the rule.

    id (str, optional):
    : Issue ID. Generated automatically from `path`, `secret` and rule UUID.

    branch (str, optional)
    : Git commit branch.

    message (str, optional)
    : Git commit message.

    author (str, optional)
    : Git commit author as `name <email>`.

    commit (str, optional)
    : Git commit hash.

    date (datetime.datetime, optional)
    : Git commit timestamp.

    Examples
    --------
    Basic usage examples

    >>> rule = Pattern(
    ...     id="bad-password-letmein",
    ...     message="Bad Password 'letmein'",
    ...     pattern="letmein",
    ...     severity="high",
    ... )
    >>> issue = Issue(
    ...     rule=rule,
    ...     path="/path/to/code.py",
    ...     line="10",
    ...     secret="letmein",
    ...     context={
    ...         "9":  "username = 'admin'",
    ...         "10": "password = 'letmein'",
    ...         "11": "response = authorize(username, password)",
    ...     },
    ... )
    >>> issue.id
    UUID('bfd860e4-2002-30dd-a1b1-24e29083c7d5')

    """

    rule: Rule = attr.ib(converter=Rule.fromany)
    path: str = attr.ib()
    line: str = attr.ib()
    secret: str = attr.ib()
    context: Context = attr.ib(converter=Context)
    id: Optional[uuid.UUID] = attr.ib()
    branch: Optional[str] = attr.ib(None)
    message: Optional[str] = attr.ib(None)
    author: Optional[str] = attr.ib(None)
    commit: Optional[str] = attr.ib(None)
    date: Optional[datetime] = attr.ib(None)

    @id.default
    def _id_default(self):
        fields = ":".join((self.path, self.secret))
        return uuid.uuid3(self.rule._uuid, fields)

    def __eq__(self, other):  # pragma: no cover
        """Override equality check to use issue IDs."""
        return (type(self) == type(other)) and (self.id == other.id)

    def __hash__(self):  # pragma: no cover
        """Override hash check to use issue ID."""
        return self.id.int

    @property
    def multiline(self) -> bool:
        """Return true if context contains multiple lines."""
        return len(self.lines) > 1

    @property
    def lines(self) -> List[int]:
        """Return context keys containing line numbers."""
        return list(self.context)

    @property
    def line_start(self) -> int:  # pragma: no cover
        """Return first context line number."""
        return self.lines[0]

    @property
    def line_end(self) -> int:
        """Return last context line number."""
        return self.lines[-1]


@attr.s
class Config(Model):
    """Config holds all configuration."""

    # search configuration
    exclude: Optional[List[Exclude]] = attr.ib(
        None,
        converter=lambda x: [Exclude.fromany(r) for r in x] if x else None,
    )
    severity: Optional[Severity] = attr.ib(Severity.LOW, converter=Severity)
    ignore_nosecret: Optional[bool] = attr.ib(IGNORE_NOSECRET)
    no_entropy: Optional[bool] = attr.ib(False)
    no_pattern: Optional[bool] = attr.ib(False)

    # source configuration
    branch: Optional[str] = attr.ib(None)
    depth: Optional[int] = attr.ib(10000)
    since: Optional[str] = attr.ib(None)
    no_current: Optional[bool] = attr.ib(False)
    no_history: Optional[bool] = attr.ib(False)

    # render configuration
    context: Optional[int] = attr.ib(0)



"""Main trufflehog3 module."""

import logging
import re

from pathlib import Path

from trufflehog3 import helper

__NAME__ = "trufflehog3"
__VERSION__ = "3.0.4"

HERE = Path(__file__).parent
STATIC_DIR = HERE / "static"
HTML_TEMPLATE_FILE = "report.html.j2"
TEXT_TEMPLATE_FILE = "report.text.j2"
DEFAULT_RULES_FILE = STATIC_DIR / "rules.yml"

DEFAULT_CONFIG_FILE = f".{__NAME__}.yml"
DEFAULT_EXCLUDE_SET = {DEFAULT_CONFIG_FILE, ".git"}

# Inline 'nosecret' comment implementation taken from semgrep:
# https://github.com/returntocorp/semgrep/blob/master/semgrep/semgrep/constants.py
NOSECRET_INLINE_RE = re.compile(
    # We're looking for items that look like this:
    # ' nosecret'
    # ' nosecret: example-pattern-id'
    # ' nosecret: pattern-id1,pattern-id2'
    # ' NOSECRET:pattern-id1, pattern-id2'
    #
    # * We do not want to capture the ': ' that follows 'nosecret'
    # * We do not care about the casing of 'nosecret'
    # * We want a comma-separated list of ids
    # * We want multi-language support, so we cannot strictly look for
    #   Python comments that begin with '# '
    #
    r" nosecret(?::[\s]?(?P<ids>([^,\s](?:[,\s]+)?)+))?",
    re.IGNORECASE,
)
IGNORE_NOSECRET = False

# global logger instance
log = helper.logger()


def set_debug(debug: bool):
    """Switch debug logging on and off.

    Examples
    --------
    Switch debug logging on and off

    >>> set_debug(True)
    >>> log.level
    10
    >>> set_debug(False)
    >>> log.level
    40

    """
    log.setLevel(logging.DEBUG if debug else logging.ERROR)



"""Render reports in all supported formats."""

import jinja2
import json as jsonlib

from collections import defaultdict
from typing import Any, Dict, Iterable, Tuple

from trufflehog3 import STATIC_DIR, HTML_TEMPLATE_FILE, TEXT_TEMPLATE_FILE
from trufflehog3.helper import Color
from trufflehog3.models import Issue, Severity, Pattern  # noqa: F401 doctest


def text(issues: Iterable[Issue]) -> str:
    """Render issues as text.

    Examples
    --------
    Basic usage examples

    >>> rule = Pattern(
    ...     id="bad-password-letmein",
    ...     message="Bad Password 'letmein'",
    ...     pattern="letmein",
    ...     severity="high",
    ... )
    >>> issue = Issue(
    ...     rule=rule,
    ...     path="/path/to/code.py",
    ...     line="10",
    ...     secret="letmein",
    ...     context={
    ...         "9":  "username = 'admin'",
    ...         "10": "password = 'letmein'",
    ...         "11": "response = authorize(username, password)",
    ...     },
    ... )
    >>> s = text([issue])

    """
    environment = jinja2.Environment(
        loader=jinja2.FileSystemLoader(STATIC_DIR),
        autoescape=False,  # no need to escape anything for plaintext format
        auto_reload=False,
    )
    template = environment.get_template(TEXT_TEMPLATE_FILE)
    return template.render(issues=sorted(issues, key=_sort_keys), color=Color)


# TODO switch to SARIF-compatible output?
def json(issues: Iterable[Issue]) -> str:
    """Render issues as JSON.

    Examples
    --------
    Basic usage examples

    >>> rule = Pattern(
    ...     id="bad-password-letmein",
    ...     message="Bad Password 'letmein'",
    ...     pattern="letmein",
    ...     severity="high",
    ... )
    >>> issue = Issue(
    ...     rule=rule,
    ...     path="/path/to/code.py",
    ...     line="10",
    ...     secret="letmein",
    ...     context={
    ...         "9":  "username = 'admin'",
    ...         "10": "password = 'letmein'",
    ...         "11": "response = authorize(username, password)",
    ...     },
    ... )
    >>> s = json([issue])

    """
    return jsonlib.dumps([i.asdict() for i in issues], indent=2, default=str)


def html(issues: Iterable[Issue]) -> str:
    """Render issues as HTML.

    Examples
    --------
    Basic usage examples

    >>> rule = Pattern(
    ...     id="bad-password-letmein",
    ...     message="Bad Password 'letmein'",
    ...     pattern="letmein",
    ...     severity="high",
    ... )
    >>> issue = Issue(
    ...     rule=rule,
    ...     path="/path/to/code.py",
    ...     line="10",
    ...     secret="letmein",
    ...     context={
    ...         "9":  "username = 'admin'",
    ...         "10": "password = 'letmein'",
    ...         "11": "response = authorize(username, password)",
    ...     },
    ... )
    >>> s = html([issue])

    """
    environment = jinja2.Environment(
        loader=jinja2.FileSystemLoader(STATIC_DIR),
        autoescape=True,
        auto_reload=False,
    )
    template = environment.get_template(HTML_TEMPLATE_FILE)
    return template.render(**_prepare_report(issues))


def _prepare_report(issues: Iterable[Issue]) -> Dict[str, Any]:
    """Split issues and rules into structs handy for HTML report building."""
    totals = defaultdict(lambda: 0)
    report = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))

    for issue in sorted(issues, key=_sort_keys):
        totals[issue.rule] += 1
        report[issue.rule][issue.path][issue.commit].append(issue)

    return dict(totals=totals, report=report)


def _sort_keys(issue) -> Tuple[Severity, str, str]:
    """Return rule severity, message and issue path for sorting."""
    return (-issue.rule.severity, issue.rule.message.lower(), issue.path)



"""Helper functions."""

import logging
import math

from typing import List, Dict


class Color:
    """Supported ANSI colors."""

    BLACK = "\x1b[30m"
    RED = "\x1b[31m"
    GREEN = "\x1b[32m"
    YELLOW = "\x1b[33m"
    BLUE = "\x1b[34m"
    MAGENTA = "\x1b[35m"
    CYAN = "\x1b[36m"
    WHITE = "\x1b[37m"
    GRAY = "\x1b[90m"

    RESET = "\x1b[0m"
    BOLD = "\x1b[1m"
    UNDERLINE = "\x1b[4m"


def colored(s: str, color: str = Color.GREEN) -> str:
    """Return ANSI-colored string."""
    return f"{color}{s}{Color.RESET}"


def logger(level: int = logging.ERROR, name="secret") -> logging.Logger:
    """Initialize and return named logger singleton.

    Note
    ----
    Level is not changed upon subsequent calls. Use log.setLevel().

    Examples
    --------
    >>> log1 = logger(level=logging.DEBUG, name="debug")
    >>> log2 = logger(level=logging.ERROR, name="debug")
    >>> log3 = logger(level=logging.ERROR, name="error")
    >>> log1 == log2
    True
    >>> log2.level  # log2 still has DEBUG level
    10
    >>> log1 == log3
    False
    >>> log2.setLevel(logging.ERROR)  # log1 and log2 now have ERROR level
    >>> log2.level
    40

    """
    log = logging.Logger.manager.loggerDict.get(name)  # type: ignore
    if log:
        return log

    logging.addLevelName(logging.DEBUG, colored("[D]", Color.CYAN))
    logging.addLevelName(logging.INFO, colored("[I]", Color.GREEN))
    logging.addLevelName(logging.WARNING, colored("[W]", Color.YELLOW))
    logging.addLevelName(logging.ERROR, colored("[E]", Color.RED))
    logging.addLevelName(logging.FATAL, colored("[F]", Color.BOLD + Color.RED))

    log = logging.getLogger(name)
    log.setLevel(level)

    formatter = logging.Formatter(
        "%(asctime)s %(levelname)s %(message)s",
        datefmt=colored("%I:%M%p", Color.GRAY),
    )

    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.DEBUG)
    stream_handler.setFormatter(formatter)
    log.addHandler(stream_handler)

    return log


def get_lines(s: str, line: int, context: int = 0) -> Dict[int, str]:
    r"""Extract lines with context from the given string.

    Return dict with lines range and the extracted lines.

    Note
    ----
    It is supposed that `line` parameter is 1-indexed.

    Examples
    --------
    Basic usage examples

    >>> s = "1\n2\n3\n4\n5"
    >>> get_lines(s, 3)
    {'3': '3'}
    >>> get_lines(s, 1, 2)
    {'1': '1', '2': '2', '3': '3'}
    >>> get_lines(s, 5, 2)
    {'3': '3', '4': '4', '5': '5'}
    >>> get_lines(s, 3, 10)
    {'1': '1', '2': '2', '3': '3', '4': '4', '5': '5'}

    """
    lines = s.splitlines()
    lower = max(0, line - context - 1)
    upper = min(len(lines), line + context)

    return {f"{i + 1}": lines[i] for i in range(lower, upper)}


def get_strings(s: str, alphabet: str, minlen: int) -> List[str]:
    """Extract substrings of given alphabet from the string.

    Examples
    --------
    Basic usage examples

    >>> get_strings("testing get_strings", "abcdefghijklmnopqrstuvwxyz", 5)
    ['testing', 'strings']
    >>> get_strings("something about deadbeef", "0123456789abcdef", 5)
    ['deadbeef']

    """
    chars = ""
    count = 0

    sub = []
    for char in s:
        if char in alphabet:
            chars += char
            count += 1
        else:
            if count >= minlen:
                sub.append(chars)

            chars = ""
            count = 0

    if count > minlen:
        sub.append(chars)

    return sub


def shannon_entropy(s: str, alphabet: str) -> float:
    """Calculate Shannon entropy for given string of alphabet characters.

    Note
    ----
    Return zero for empty string.

    Examples
    --------
    Basic usage examples

    >>> shannon_entropy("", "abcdefghijklmnopqrstuvwxyz")
    0.0
    >>> shannon_entropy("abcd", "abcdefghijklmnopqrstuvwxyz")
    2.0
    >>> shannon_entropy("1234abcd", "0123456789abcdef")
    3.0

    """
    entropy = 0.0
    if not s:
        return entropy

    for x in alphabet:
        px = float(s.count(x)) / len(s)
        if px > 0:
            entropy += -px * math.log(px, 2)

    return entropy



"""Distribution settings."""

from pathlib import Path
from setuptools import setup

from trufflehog3 import __NAME__, __VERSION__

HERE = Path(__file__).parent
README = (HERE / "README.md").read_text()
REQUIREMENTS = (HERE / "requirements.txt").read_text().splitlines()

setup(
    name=__NAME__,
    version=__VERSION__,
    packages=[__NAME__],
    license="GNU",
    description="Find secrets in your codebase",
    long_description=README,
    long_description_content_type="text/markdown",
    url=f"https://github.com/feeltheajf/{__NAME__}",
    author="Ilya Radostev",
    author_email="feeltheajf@gmail.com",
    include_package_data=True,
    python_requires=">=3.7",
    install_requires=REQUIREMENTS,
    entry_points={"console_scripts": [f"{__NAME__} = {__NAME__}.cli:run"]},
)
